/*
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *		  ####    #####  #    #  ######  ######
 *		 #          #    #    #  #       #
 *		  ####      #    #    #  #####   #####
 *		      #     #    #    #  #       #
 *		 #    #     #    #    #  #       #
 *		  ####      #     ####   #       #
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Author:   Goran Larsson
 *
 * Date:     6-jul-1987
 *
 * Synopsis: Stuff fills the streamer with data and reports the number
 *           of bytes written. Use stuff to determine capacity of tape
 *           for use by tar and turbo.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Usage:
 *
 *	stuff <device>
 *
 *	<device> is a special file (device) (streamer).
 *
 *	Ignore error messages produced by device driver.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Notes:
 *
 * DMASIZE: This is the maximum DMA size allowed for the system used.
 *          For ABC1600 this is 32kb if the buffer is allocated
 *          contiguous in PHYSICAL memory, but 2kb otherwise. The DMA
 *          channels has only mapper, not MMU.
 *
 * MAXSIZE: This is the initial amount of RAM asked for as buffer.
 *
 * MINSIZE: This is the minimum amount of RAM accepted as buffer.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */

#include <stdio.h>

/*
 *
 * open modes
 *
 */

#define READ 0
#define WRITE 1
#define UPDATE 2

/*
 *
 * logical values
 *
 */

#define FALSE 0
#define TRUE -1

/*
 *
 * machine dependant figures
 *
 */

#define DMASIZE (2*1024)
#define MAXSIZE (300*1024)
#define MINSIZE (60*1024)

#define Kilo 1024
#define Meg (1024*1024)

#define STUFF 0xE5

/*
 *
 * externals
 *
 */

extern int errno;
extern char *malloc();		/* used for storage allocation */

/*********
*        *
*  main  *
*        *
*********/

main ( argc, argv )
int argc;
char *argv[];

{
int fdout;			/* file descriptor */

long bigbufsz, bp, bm;		/* buffer pointers */
int m;
unsigned int rc;
char *bbuf;

int i, done;			/* flags and counter */

long unsigned int bytes_total;	/* total number of bytes buffered */
long unsigned int w_total;	/* total number of writes */
long unsigned int wb_total;	/* total number of buffered writes */

	/*
	 *
	 * parse command line
	 *
	 */

	if ( argc == 2 ) {
		fdout = open ( argv[1], WRITE );
	} else {
		fprintf ( stderr,"stuff: need <device>\n" );
		exit ( 1 );
	}

	if ( fdout < 0 ) {
		fprintf ( stderr,"stuff: can't open %s\n", argv[1] );
		exit ( 1 );
	}

	/*
	 *
	 * allocate buffer
	 *
	 */

	bigbufsz = MAXSIZE;
	while ( 1 ) {
		if ( ( bbuf = malloc ( bigbufsz ) ) == NULL ) {
			if ( bigbufsz < MINSIZE ) {
				fprintf ( stderr,"stuff: no buffer\n");
				exit ( 1 );
			} else {
				bigbufsz /= 2;
			}
		} else {
			break;
		}
	}
	if ( bigbufsz != MAXSIZE ) {
		fprintf ( stderr,"stuff: reduced buffer, %ld\n", bigbufsz );
	}

	/*
	 *
	 * clear statistics collectors
	 *
	 */

	bytes_total = (long) 0;
	w_total = (long) 0;
	wb_total = (long) 0;

	/*
	 *
	 * fill buffer
	 *
	 */

	for ( bp = (long) 0; bp < bigbufsz; bp++ ) {
		bbuf[bp] = STUFF;
	}

	/*
	 *
	 * loop here until streamer yells
	 *
	 */

	done = FALSE;
	while ( !done ) {

		/*
		 *
		 * stream buffer
		 *
		 */

		bm = bp;
		bp = 0;
		while ( bp < bm ) {
			if ( ( bm - bp ) < DMASIZE ) {
				rc = bm - bp;
			} else {
				rc = DMASIZE;
			}
			m = write ( fdout, &bbuf[bp], rc );
			w_total++;
			if ( errno ) {

				/*
				 * If errno is 12 then ABCenix was unable
				 * to lock memory. The buffer was allocated
				 * but did not fit in physical memory. We
				 * ought to do something about it!
				 */

				sleep ( 5 );	/* sleep until log written */

				fprintf ( stderr,"\nstuff: write failed\n");
				fprintf ( stderr,"stuff: errno=%d\n", errno);
				done = TRUE;
			}
			bp += m;
			bytes_total += m;
			if ( done ) {
				break;
			}
		}
		wb_total++;
/***********
		if ( ( bytes_total % Meg ) == 0 ) {
			fprintf (
				stderr,
				"stuff: current: %ld bytes, %.0fk, %.2fM\n",
				bytes_total,
				( (float) bytes_total ) / (float) Kilo,
				( (float) bytes_total ) / (float) Meg
			);
		}
***********/
	}

	/*
	 *
	 * statistics, free buffer, and exit
	 *
	 */

	fprintf ( stderr,"\nstuff: --------------- final ---------------\n" );
	fprintf (
		stderr,
		"stuff: final: %ld bytes, %.0fk, %.2fM\n",
		bytes_total,
		( (float) bytes_total ) / (float) Kilo,
		( (float) bytes_total ) / (float) Meg
	);
	fprintf ( stderr,"stuff: final: writes %ld/%ld\n", w_total, wb_total );
	fprintf ( stderr,"stuff: -------------------------------------\n" );

	fprintf ( stderr,"\nstuff: tape rewinding...\n" );

	free ( bbuf );
	exit ( 0 );
}
