/*
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *		 #####  #    #  #####   #####    ####
 *		   #    #    #  #    #  #    #  #    #
 *		   #    #    #  #    #  #####   #    #
 *		   #    #    #  #####   #    #  #    #
 *		   #    #    #  #   #   #    #  #    #
 *		   #     ####   #    #  #####    ####
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Author:   Goran Larsson
 *
 * Date:     5-jul-1987
 *
 * Synopsis: Turbo runs as a start or an end of a pipe and does the
 *           I/O to a device (usually a streamer). It gathers data
 *           in a big buffer (256K) and does I/O at a rate that keeps
 *           a streamer streaming.
 *           This will effectively prevent "shoeshining" the R/W
 *           heads of the streamer.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Graphical description of performance with and without turbo:
 *
 * > streamer forward
 * < streamer backward
 * - streamer idle
 * X tar executing
 * _ tar waiting until write to streamer has completed
 * x tar executing slowly due to I/O competition until pipe is full,
 *   then waiting until turbo starts reading pipe again
 *
 * tar operations on streamer:
 *
 * XXX____XXX____XXX____XXX____XXX____XXX____XXX____XXX____XXX...
 * --->>>><<<>>>><<<>>>><<<>>>><<<>>>><<<>>>><<<>>>><<<>>>><<<...
 *
 * tar/turbo operations on streamer:
 *
 * XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXxxxxxxxxxxxxxxxxXXXXXXXXX...
 * ---------------------------------->>>>>>>>>>>>>>>><<<------...
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Real life example:
 *
 * 1) tar save of 10Mb from 70Mb winchester to streamer
 *
 *      tar cf /dev/st0 /
 *
 *    takes 42 minutes real time. Streamer is shoeshining for 42 minutes.
 *
 * 2) tar/turbo save of 10Mb from 70Mb winchester to streamer
 *
 *      tar cf - / | turbo -o /dev/st0
 *
 *    takes 13 minutes real time. Streamer streaming for 3 minutes.
 *    (tar is slow!)
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Usage:
 *
 *	turbo [-][i|o][v] <device>
 *
 *	i  Turbo at start of pipe.
 *	   Input from device, output to stdout (pipe).
 *
 *	o  Turbo at end of pipe.
 *	   Input from stdin (pipe), output to device.
 *
 *	v  Verbose mode. Print statistics (to stderr).
 *
 *	<device> is usually a special file (device) (streamer).
 *
 *	If i and o not specified, then input comes from std input
 *	and output goes to std output. This may be in the middle of
 *	a pipe.
 *
 *	Std input and std output may be redirected.
 *
 *	<device> is opened for UPDATE in o-mode, and READ in i-mode.
 *
 *	When using turbo with tar, tar can only create (c), extract (x),
 *	and test (t) archives. Functions append (r) and update (u) can
 *	not be used when tar is part of a pipe.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Examples:
 *
 *	Boost tar input from streamer
 *
 *			turbo -i /dev/st0 | tar xf -
 *
 *	Boost tar input from streamer with statistics
 *
 *			turbo -iv /dev/st0 | tar xf -
 *
 *	Boost tar output to streamer
 *
 *			tar cf - / | turbo -o /dev/st0
 *
 *	Boost tar output to streamer with statistics
 *
 *			tar cf - / | turbo -ov /dev/st0
 *
 *	Boost tar output to streamer with statistics and size control
 *	for multivolume archives. Tar can not detect tape end when
 *	piping. This means that we must use (k) option to let tar know
 *	the size of the tape when we expect that one streamer is to
 *	little for the data we are about to archive. In this example
 *	the tape is a DC600A wich will store 60 Mbyte = 61440 kbyte.
 *
 *			tar cfk - 61440 / | turbo -ov /dev/st0
 *
 *	Copy file
 *
 *			turbo <fromfile >tofile
 *
 *	Copy file with statistics
 *
 *			turbo -v <fromfile >tofile
 *
 *	Copy device to file
 *
 *			turbo -i /dev/mf2 >floppyimagefile
 *
 *	Copy file to device
 *
 *			turbo -o /dev/mf2 <floppyimagefile
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Notes:
 *
 * DMASIZE: This is the maximum DMA size allowed for the system used.
 *          For ABC1600 this is 32kb if the buffer is allocated
 *          contiguous in PHYSICAL memory, but 2kb otherwise. The DMA
 *          channels has only mapper, not MMU.
 *
 * MAXSIZE: This is the initial amount of RAM asked for as buffer.
 *
 * MINSIZE: This is the minimum amount of RAM accepted as buffer.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */

#include <stdio.h>

/*
 *
 * predefined file descriptors
 *
 */

#define STDIN 0
#define STDOUT 1
#define STDERR 2

/*
 *
 * open modes
 *
 */

#define READ 0
#define WRITE 1
#define UPDATE 2

/*
 *
 * logical values
 *
 */

#define FALSE 0
#define TRUE -1

/*
 *
 * machine dependant figures
 *
 */

#define DMASIZE (2*1024)		/* TOO SMALL! */
#define MAXSIZE (300*1024)
#define MINSIZE (60*1024)

/*  #define DMASIZE (32*1024)		/* THIS WORKS SOMETIMES */
 *  #define MAXSIZE (256*1024)
 *  #define MINSIZE (64*1024)
 */

/*
 *
 * externals
 *
 */

extern int errno;
extern char *malloc();		/* used for storage allocation */

/*********
*        *
*  main  *
*        *
*********/

main ( argc, argv )
int argc;
char *argv[];

{
int fdin, fdout;		/* file descriptors */

long bigbufsz, bp, bm;		/* buffer pointers */
int m;
unsigned int rc;
char *bbuf;

int i, verbose, done;		/* flags and counter */

long unsigned int bytes_total;	/* total number of bytes buffered */
long unsigned int r_total;	/* total number of reads */
long unsigned int w_total;	/* total number of writes */
long unsigned int rb_total;	/* total number of buffered reads */
long unsigned int wb_total;	/* total number of buffered writes */

	/*
	 *
	 * scan argument vector
	 *
	 */

	verbose = FALSE;
	done = FALSE;

	if ( argc == 1 ) {
		fdin = STDIN;
		fdout = STDOUT;
		done = TRUE;
	}
	if ( argc == 2 ) {
		fdin = STDIN;
		fdout = STDOUT;
		for ( i = 0; i < strlen ( argv[1] ); i++ ) {
			switch ( argv[1][i] ) {
				case '-':
					break;
				case 'i':
					break;
				case 'o':
					break;
				case 'v':
					verbose = TRUE;
					break;
				default:
		   			fprintf (
						stderr,
						"turbo: bad switch %c\n",
						argv[1][i]
					);
					exit ( 1 );
					break;
			}
		}
		done = TRUE;
	}
	if ( argc == 3 ) {
		for ( i = 0; i < strlen ( argv[1] ); i++ ) {
			switch ( argv[1][i] ) {
				case '-':
					break;
				case 'i':
					fdin = open ( argv[2], READ );
					fdout = STDOUT;
					break;
				case 'o':
					fdin = STDIN;
					fdout = open ( argv[2], UPDATE );
					break;
				case 'v':
					verbose = TRUE;
					break;
				default:
		   			fprintf (
						stderr,
						"turbo: bad switch %c\n",
						argv[1][i]
					);
					exit ( 1 );
					break;
			}
		}
		done = TRUE;
	}
	if ( !done ) {
		fprintf ( stderr,"turbo: bad command format\n" );
		exit ( 1 );
	}
	if ( ( fdin < 0 ) || ( fdout < 0 ) ) {
		fprintf ( stderr,"turbo: can't open %s\n", argv[2] );
		exit ( 1 );
	}

	/*
	 *
	 * allocate buffer
	 *
	 */

	bigbufsz = MAXSIZE;
	while ( 1 ) {
		if ( ( bbuf = malloc ( bigbufsz ) ) == NULL ) {
			if ( bigbufsz < MINSIZE ) {
				fprintf ( stderr,"turbo: no buffer\n");
				exit ( 1 );
			} else {
				bigbufsz /= 2;
			}
		} else {
			break;
		}
	}
	if ( bigbufsz != MAXSIZE ) {
		fprintf ( stderr,"turbo: reduced buffer, %ld\n", bigbufsz );
	}

	/*
	 *
	 * clear statistics collectors
	 *
	 */

	bytes_total = (long) 0;
	r_total = (long) 0;
	w_total = (long) 0;
	rb_total = (long) 0;
	wb_total = (long) 0;

	/*
	 *
	 * loop here until eof
	 *
	 */

	done = FALSE;
	while ( !done ) {

		/*
		 *
		 * fill buffer
		 *
		 */

		bp = 0;
		while ( bp < bigbufsz ) {
			if ( ( bigbufsz - bp ) < DMASIZE ) {
				rc = bigbufsz - bp;
			} else {
				rc = DMASIZE;
			}
			m = read ( fdin, &bbuf[bp], rc );
			r_total++;
			if ( m < 0 ) {

				/*
				 * If errno is 12 then ABCenix was unable
				 * to lock memory. The buffer was allocated
				 * but did not fit in physical memory. We
				 * ought to do something about it!
				 */

				sleep ( 5 );	/* sleep until log written */

				fprintf ( stderr,"turbo: read failed\n" );
				fprintf ( stderr,"      errno=%d\n", errno );
				free ( bbuf );
				exit ( 1 );
			}
			if ( m == 0 ) {
				done = TRUE;
				break;		/* break out of read at eof */
			}
			bp += m;
			bytes_total += m;
		}
		rb_total++;

		/*
		 *
		 * empty buffer
		 *
		 */

		bm = bp;
		bp = 0;
		while ( bp < bm ) {
			if ( ( bm - bp ) < DMASIZE ) {
				rc = bm - bp;
			} else {
				rc = DMASIZE;
			}
			m = write ( fdout, &bbuf[bp], rc );
			w_total++;
			if ( m != rc ) {

				/*
				 * If errno is 12 then ABCenix was unable
				 * to lock memory. The buffer was allocated
				 * but did not fit in physical memory. We
				 * ought to do something about it!
				 */

				sleep ( 5 );	/* sleep until log written */

				fprintf ( stderr,"turbo: write failed\n");
				fprintf ( stderr,"      errno=%d\n", errno);
				fprintf ( stderr,"    written=%d\n", m);
				free ( bbuf );
				exit ( 1 );
			}
			bp += m;
		}
		wb_total++;
	}

	/*
	 *
	 * statistics, free buffer, and exit
	 *
	 */

	if ( verbose ) {
		fprintf ( stderr,"turbo: bytes %ld\n", bytes_total );
		fprintf ( stderr,"turbo: read  %ld/%ld\n", r_total, rb_total );
		fprintf ( stderr,"turbo: write %ld/%ld\n", w_total, wb_total );
	}

	free ( bbuf );
	exit ( 0 );
}
